#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

xhtml1.py

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

A modified XHTML writer.

"""

from __future__ import unicode_literals

# FIXME: move hr.pb out of sections

import six

from docutils import nodes
from docutils.writers import html4css1

from libgutenberg.Logger import info, debug, warning, error
from libgutenberg.GutenbergGlobals import xpath

from ebookmaker.mydocutils.transforms import parts
from ebookmaker.mydocutils import writers

class InheritTransform (parts.InheritTransform):
    pass


class Writer (html4css1.Writer):
    """ XHTML 1 Writer. """

    def __init__ (self):
        html4css1.Writer.__init__ (self)
        self.translator_class = Translator

    def fixup_xhtml (self, xhtml):
        # remove bad attribute
        attrs_to_remove = [('*', 'classes'),]
        for (tag, attr) in attrs_to_remove:
            for elem in xpath(xhtml, f"//xhtml:{tag}[@{attr}]"):
                del elem.attrib[attr]
        
        return xhtml


    def get_transforms (self):
        tfs = html4css1.Writer.get_transforms (self)
        return tfs + [parts.TextNodeWrapper, InheritTransform]


class Translator (writers.Translator, html4css1.HTMLTranslator):
    """ XHTML 1 Translator. """

    htmlspecialchars = {
        ord('&'): '&amp;',
        ord('<'): '&lt;',
        ord('>'): '&gt;',
        ord('"'): '&#x22;',
        0xa0:     '&#xa0;',
        }

    head_prefix_template = ('<html xmlns="http://www.w3.org/1999/xhtml"'
                            ' xml:lang="%(lang)s">\n<head>\n')

    def __init__ (self, document):
        writers.Translator.__init__ (self, document)
        html4css1.HTMLTranslator.__init__ (self, document)
        self.init_css ()
        self.imgoff = None

    def init_css (self):
        for css_file in ('rst2all.css', 'rst2html.css'):
            self.head.append ('<style type="text/css">\n%s</style>\n' %
                              self.encode (self.read_css (css_file)))

    def read_css (self, css_file):
        if self.document.settings.get_resource and six.callable (
            self.document.settings.get_resource):
            return self.document.settings.get_resource ('mydocutils.writers', css_file)
        return None

    def encode (self, text):
        """ Encode special html characters.

        Tweak: Use numeric entities that work in HTML and XHTML. """

        return six.text_type (text).translate (self.htmlspecialchars)

    def visit_Text (self, node):
        text = node.astext ()
        encoded = self.encode (text)
        if self.in_mailto and self.settings.cloak_email_addresses:
            encoded = self.cloak_email (encoded)

        if hasattr (node, 'attributes') and 'white-space-pre-line' in node.attributes['classes']:
            for line in encoded.splitlines (True):
                self.body.append (line)
                if '\n' in line:
                    self.body.append ('<br />')
        else:
            self.body.append (encoded)

    def starttag (self, node, tagname, suffix='\n', empty=0, **attributes):
        """
        Tweak is: implement custom html attributes.

        Tweak is: original starttag produces side-effects on the node
        it operates on: ie. it changes the `classes` dict. Fix that.

        Construct and return a start tag given a node (id & class attributes
        are extracted), tag name, and optional attributes.

        """

        tagname = tagname.lower ()
        prefix = []
        atts = {}

        for (name, value) in attributes.items ():
            atts[name.lower()] = value
        if 'html_attributes' in node:
            atts.update (node['html_attributes'])

        classes = node.get ('classes', [])[:] # fix: make a copy

        if 'class' in atts:
            class_ = atts['class']
            if isinstance (class_, six.string_types):
                class_ = [class_]
            classes.extend (class_)
            del atts['class']

        # language hack
        for c in classes:
            if c.startswith ('language-'):
                atts['xml:lang'] = c[9:]

        styles = node.get ('styles', [])[:] # fix: make a copy

        if 'style' in atts:
            style = atts['style']
            if isinstance (style, six.string_types):
                style = [style]
            styles.extend (style)
            del atts['style']
        styles = [s.strip (' ;') for s in styles]

        assert 'id' not in atts
        ids = node.get ('ids', [])
        if 'ids' in atts:
            ids.extend (atts['ids'])
            del atts['ids']
        if ids:
            atts['id'] = ids[0]
            for id_ in ids[1:]:
                # Add empty "span" elements for additional IDs.  Note
                # that we cannot use empty "a" elements because there
                # may be targets inside of references, but nested "a"
                # elements aren't allowed in XHTML (even if they do
                # not all have a "href" attribute).
                if empty:
                    # Empty tag.  Insert target right in front of element.
                    prefix.append('<span id="%s"></span>' % id_)
                else:
                    # Non-empty tag.  Place the auxiliary <span> tag
                    # *inside* the element, as the first child.
                    suffix += '<span id="%s"></span>' % id_

        parts = []
        if classes:
            parts.append ('class="%s"' % ' '.join (sorted (set (classes))))
        if styles:
            parts.append ('style="%s"' % '; '.join (sorted (set (styles))))
        for name, value in sorted (atts.items ()):
            parts.append ('%s="%s"' % (name.lower (), self.attval (six.text_type (value))))

        infix = ' /' if empty else ''

        return ''.join(prefix) + '<%s %s%s>' % (tagname, ' '.join(parts), infix) + suffix


    def set_first_last (self, node):
        """ Set class 'first' on first child, 'last' on last child. """
        self.set_class_on_child (node, 'first', 0)
        self.set_class_on_child (node, 'last', -1)

    def depart_document (self, node):
        link_data = {'coverpage':('icon', 'image/x-cover'), 'page-images':('alt', 'text/*')}
        for key, values in node.meta_block.items ():
            for val in values:
                val = self.attval (six.text_type (val))
                if key in link_data:
                    key, type_ = link_data[key]        
                    self.head.append('<link rel="%s" href="%s" type= "%s" />\n' % (key, val, type_))
                else:
                    self.head.append ('<meta name="%s" content="%s" />\n' % (key, val))

        html4css1.HTMLTranslator.depart_document (self, node)

    def visit_meta (self, node):
        # not used any more
        pass

    def visit_title (self, node):
        if isinstance (node.parent, nodes.table):
            self.body.append (self.starttag (node, 'caption'))
            if node.hasattr('refid'):
                atts = {
                    'class': 'toc-backref',
                    'href': '#' + node['refid'],
                    }
                self.body.append (self.starttag ({}, 'a', **atts))
                self.context.append ('</a></caption>\n')
            else:
                self.context.append ('</caption>\n')
        else:
            html4css1.HTMLTranslator.visit_title (self, node)


    def depart_title (self, node):
        html4css1.HTMLTranslator.depart_title (self, node)


    def visit_subtitle (self, node):
        """ Tweak is: use <p> instead of <h?> """

        if isinstance (node.parent, (nodes.document, nodes.section, nodes.topic)):
            self.body.append (self.starttag (node, 'p'))
            self.context.append ('</p>\n')
            if isinstance (node.parent, nodes.document):
                self.in_document_title = len (self.body)
            return

        html4css1.HTMLTranslator.visit_subtitle (self, node)


    def depart_subtitle (self, node):
        html4css1.HTMLTranslator.depart_subtitle (self, node)


    def visit_caption (self, node):
        """ Tweak is: use <div> instead of <p> """
        self.body.append (self.starttag (node, 'div', CLASS='caption'))

        atts = {}
        if node.hasattr ('refid'):
            atts['class'] = 'toc-backref'
            atts['href'] = '#' + node['refid']
            self.body.append (self.starttag ({}, 'a', **atts))


    def depart_caption (self, node):
        """ Tweak is: use <div> instead of <p> """
        if node.hasattr ('refid'):
            self.body.append ('</a>')
        self.body.append ('</div>\n')


    def visit_line (self, node):
        """ Tweak is: fix empty lines on ADE. """

        self.body.append (self.starttag (node, 'div', '', CLASS='line'))
        if not len (node):
            self.body.append (' ') # U+00A0 nbsp


    def visit_line_block (self, node):
        """ Tweak is: noindent if centered or right-aligned. """

        extraclasses = ['line-block']
        if isinstance (node.parent, nodes.line_block):
            extraclasses.append ('inner')
        else:
            extraclasses.append ('outermost')
        if 'center' in node['classes'] or 'right' in node['classes']:
            extraclasses.append ('noindent')
        self.body.append (self.starttag (node, 'div', **{ 'class': extraclasses }))


    def visit_table (self, node):
        self.context.append (self.compact_p)
        self.compact_p = True

        node['styles'] = []
        node['html_attributes'] = {}

        pass_1 = writers.TablePass1 (self.document)
        node.walkabout (pass_1)

        options = { 'class': ['table'] }
        if 'align' in node:
            options['class'].append ('align-%s' % node['align'])
        if 'summary' in node:
            options['summary'] = node['summary']
        if 'table' in node['hrules']:
            options['class'].append ('hrules-table')
        if 'rows' in node['hrules']:
            options['class'].append ('hrules-rows')
        if 'table' in node['vrules']:
            options['class'].append ('vrules-table')
        if 'columns' in node['vrules']:
            options['class'].append ('vrules-columns')
        self.calc_centering_style (node)
        self.body.append (self.starttag (node, 'table', **options))

    def depart_table (self, node):
        self.compact_p = self.context.pop ()
        self.body.append ('</table>\n')

    def visit_thead (self, node):
        self.set_first_last (node)
        html4css1.HTMLTranslator.visit_thead (self, node)

    def visit_tbody (self, node):
        self.set_first_last (node)
        html4css1.HTMLTranslator.visit_tbody (self, node)

    def visit_entry (self, node):
        """ Tweak is: put alignment on element. """

        if 'vspan' in node:
            # HTML spans natively
            raise nodes.SkipNode

        node['styles'] = []
        align = node.colspecs[0].get ('align', 'left')
        if align != 'left':
            node['styles'].append ("text-align: %s" % align)
        valign = node.colspecs[0].get ('valign', 'middle')
        if valign != 'middle':
            node['styles'].append ("vertical-align: %s" % valign)

        html4css1.HTMLTranslator.visit_entry (self, node)

    def visit_target (self, node):
        if 'pageno' in node['classes'] or 'lineno' in node['classes']:
            options = { 'class': 'target' }
            # the ' ' avoids an empty span which will cause trouble if
            # parsed by a HTML parser.
            # FIXME: move pageno out of caption

            # bump image offset
            if self.imgoff:
                self.imgoff += 1
            if 'imgoff' in node:
                self.imgoff = int (node['imgoff'])

            if self.imgoff and 'page-images' in self.document.meta_block:
                href = self.document.meta_block['page-images'][0]
                href = href.format (page = self.imgoff)
                node['html_attributes']['href'] = href
                if 'invisible' in node['classes']:
                    node['html_attributes']['title'] = ' '
                    del node['classes'][node['classes'].index ('invisible')]
                self.body.append (self.starttag (node, 'a', ' ', **options))
                self.context.append('</a>')
            else:
                self.body.append (self.starttag (node, 'span', ' ', **options))
                self.context.append('</span>')
            return
        html4css1.HTMLTranslator.visit_target (self, node)


    def visit_page (self, node):
        if 'vspace' in node['classes']:
            node['styles'] = [ 'height: %dem' % node['length'] ]
        self.body.append (self.starttag (node, 'div'))

    def depart_page (self, node):
        self.body.append ('</div>\n')

        #c = node['classes']
        #if 'clearpage' in c or 'cleardoublepage' in c:
        #    self.body.append (self.starttag (node, 'hr', empty = 1))

    def visit_newline (self, node):
        if 'white-space-pre-line' in node['classes']:
            self.body.append ('<br />\n')
        else:
            self.body.append ('\n')

    def depart_newline (self, node):
        pass

    def visit_inline (self, node, extra_classes = []):
        options = {}
        if 'dropcap' in node['classes']:
            node['styles'] = [ "font-size: %.2fem" % (
                float (node.get ('lines', '2')) * 1.5) ]
        self.body.append (self.starttag (node, 'span', '', **options))


    def calc_centering_style (self, node):
        """
        To be overwritten in derived writers.

        :align: center has not the same semantics as :class: center.
        Former centers the block, eg. the whole table, latter centers
        the text, eg, the text in every table cell.

        :align: is supposed to work on blocks. It floats or centers
        a block.

            `:align: center`
                Used on image: centers image
                Used on figure: centers image and caption
                Used on table: centers table and caption

        """

        styles = node['styles']
        if 'width' in node and node ['width'] != 'image':
            styles.append ('width: %s' % node ['width'])


    def visit_image (self, node):
        node['styles'] = []
        node['html_attributes'] = {}
        if not node.get ('alt'):
            node['alt'] = ' '

        if 'dropcap' in node['classes']:
            node['height'] = "%.2fem" % (float (node.get ('lines', '2')) * 1.2)

        # check if we are a block image.
        # see class `TextElement` in `docutils.nodes`.
        if not isinstance (node.parent, nodes.TextElement):
            self.calc_centering_style (node)
            node['styles'].append ('display: block')

        html4css1.HTMLTranslator.visit_image (self, node)


    def depart_image (self, node):
        html4css1.HTMLTranslator.depart_image (self, node)


    def visit_figure (self, node):
        node['styles'] = []
        node['html_attributes'] = {}
        options = {}

        class_ = ['figure']
        if 'align' in node:
            class_.append ('align-' + node['align'])
        options['class'] = class_

        self.calc_centering_style (node)

        self.body.append (self.starttag (node, 'div', **options))


    def visit_block_quote (self, node):
        self.body.append (self.starttag (node, 'blockquote'))
        self.body.append ('<div>\n')

    def depart_block_quote(self, node):
        self.body.append ('</div>\n</blockquote>\n')


    def visit_footnote_group (self, node):
        self.body.append (self.starttag (node, 'div', CLASS='footnote-group'))

    def depart_footnote_group (self, node):
        self.body.append ('</div>\n')


    def visit_transition (self, node):
        self.body.append (self.starttag (node, 'div', CLASS='transition'))

    def depart_transition (self, node):
        self.body.append ('</div>\n')


    def visit_attribution(self, node):
        self.body.append (self.starttag (node, 'div', CLASS='attribution'))

    def depart_attribution(self, node):
        self.body.append ('</div>\n')



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

Packager package

Copyright 2009-2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Base class for Packager modules.

"""


import os.path
import gzip
import zipfile

from pkg_resources import resource_listdir  # pylint: disable=E0611

from libgutenberg.Logger import debug, info, warning, error
import libgutenberg.GutenbergGlobals as gg

GZIP_EXTENSION = '.gzip'

class BasePackager (object):
    """
    Base class for Packagers.

    """

    def __init__ (self):
        self.path_name_ext = None
        self.path = None
        self.name = None
        self.ext = None


    def setup (self, job):
        """ Setup """

        self.path_name_ext = os.path.join(os.path.abspath(job.outputdir), job.outputfile)
        self.path, name = os.path.split (self.path_name_ext)
        self.name, self.ext = os.path.splitext (name)


    def package (self, job):
        """ Package files. """
        pass


class OneFileGzipPackager (BasePackager):
    """ Gzips one file. """

    def package (self, job):
        self.setup (job)
        filename = self.path_name_ext
        gzfilename = filename + GZIP_EXTENSION

        try:
            info ('Creating Gzip file: %s' % gzfilename)
            info ('  Adding file: %s' % filename)
            with open (filename, 'rb') as fp:
                with gzip.open (gzfilename, 'wb') as fpgz:
                    fpgz.writelines (fp)
            info ('Done Zip file: %s' % gzfilename)
        except IOError as what:
            error (what)


class ZipPackager (BasePackager):
    """ Packages a zip file. """

    @staticmethod
    def create (zipfilename):
        """ Create a zip file. """

        info ('Creating Zip file: %s' % zipfilename)
        return  zipfile.ZipFile (zipfilename, 'w', zipfile.ZIP_DEFLATED)


    @staticmethod
    def add (zip_, filename, memberfilename):
        """ Add one file to the zip. """

        try:
            os.stat (filename)
            dummy_name, ext = os.path.splitext (filename)
            info ('  Adding file: %s as %s' % (filename, memberfilename))
            zip_.write (filename, memberfilename,
                        zipfile.ZIP_STORED if ext in ['.zip', '.png']
                        else zipfile.ZIP_DEFLATED)
        except OSError:
            warning ('ZipPackager: Cannot add file %s', filename)


class OneFileZipPackager (ZipPackager):
    """ Packages one file in zip of the same name. """

    def package (self, job):
        self.setup (job)
        filename = self.path_name_ext
        zipfilename = os.path.join (self.path, self.name) + '.zip'
        memberfilename = self.name + self.ext

        zip_ = self.create (zipfilename)
        self.add (zip_, filename, memberfilename)
        zip_.close ()

        info ('Done Zip file: %s' % zipfilename)


class HTMLishPackager (ZipPackager):
    """ Package a file with images. """

    def package (self, job):
        self.setup (job)

        try:
            aux_file_list = list (job.spider.aux_file_iter ())
        except AttributeError:
            aux_file_list = []

        filename = job.outputfile
        zipfilename = os.path.join (self.path, self.name) + '.zip'
        memberfilename = os.path.join (self.name, self.name) + self.ext

        zip_ = self.create (zipfilename)
        self.add (zip_, filename, memberfilename)

        # now images
        for url in aux_file_list:
            rel_url = gg.make_url_relative (job.base_url, url)
            filename = os.path.join (self.path, rel_url)
            memberfilename = os.path.join (self.name, rel_url)
            self.add (zip_, filename, memberfilename)

        zip_.close ()
        info ('Done Zip file: %s' % zipfilename)


class PackagerFactory (object):
    """ Implements Factory pattern for packagers. """

    packagers = {}

    @staticmethod
    def mk_key (type_, format_):
        """ Make a key for the packager map. """

        return (type_ or '') + '/' + format_


    @classmethod
    def load_packagers (cls):
        """ Load the packagers in the packagers directory. """

        for fn in resource_listdir ('ebookmaker.packagers', ''):
            modulename, ext = os.path.splitext (fn)
            if ext == '.py':
                if modulename.endswith ('Packager'):
                    module = __import__ ('ebookmaker.packagers.' + modulename,
                                         fromlist = [modulename])
                    debug ("Loading packager type: %s from module: %s for formats: %s" % (
                        module.TYPE, modulename, ', '.join (module.FORMATS)))
                    for format_ in module.FORMATS:
                        cls.packagers[cls.mk_key (module.TYPE, format_)] = module

        return cls.packagers.keys ()


    @classmethod
    def unload_packagers (cls):
        """ Unload packager modules. """

        for k in list (cls.packagers.keys ()):
            del cls.packagers[k]


    @classmethod
    def create (cls, type_, format_):
        """ Create a packager for format. """

        module = cls.packagers.get (cls.mk_key (type_, format_))
        if module:
            return module.Packager ()
        return None



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

boilerplate.py

Copyright 2022 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

This module finds Project Gutenberg boilerplate and, if found puts it into 3 section elements:

pg_header
    usually a title and license declaration
    sometimes, title, book number, release date, authors, language, encoding, credits
    when detected, metadata will be parsed and enclosed in a pg_metadata_raw sub-section

pg_footer
    usually the license
    
pg_smallprint
    on older books, this will contain license-ish language and other material. it's usually found 
    at the top of the text, and is often comically dated. 


BeautifulSoup is used for its superior "scraping" tools.

"""

import copy
import re

from bs4 import BeautifulSoup
import soupsieve as sv

from libgutenberg.Logger import critical, info, debug, warning, error


TOP_MARKERS = [
    re.compile(r"\*+ ?START OF TH(E|IS) PROJECT GUTENBERG", re.I),
]
BOTTOM_MARKERS = [
    re.compile(r"\** ?END OF TH(E|IS) PROJECT GUTENBERG", re.I),
    re.compile(r"\** ?Ende dieses Projekt Gutenberg", re.I),
    re.compile(r"\** ?END OF PROJECT GUTENBERG", re.I),
    re.compile(r"\** ?End of the Project Gutenberg", re.I),
]
SMALLPRINT_MARKERS = [
    re.compile(r"\** ?END\*? ?THE SMALL PRINT", re.I),
    re.compile(r"\**END THE SMALL PRINT", re.I),
    re.compile(r"\** ?These \w+ Were Prepared By Thousands", re.I),
]

def prune(root, divider, after=True):
    ''' prune parts of the root element before or after a divider element '''
    def next_or_prev(el, after=True):
         return el.next_sibling if after else el.previous_sibling

    def after_or_before(el, after=True):
         return list(el.next_siblings) if after else list(el.previous_siblings)

    dividers = [divider] + list(divider.parents)
    keep = False
    for elem in dividers:
        if elem is root:
            break
        has_sibling = bool(next_or_prev(elem, after=not after))
        for sibling in after_or_before(elem, after=after):
            sibling.extract()
        keep = has_sibling or keep

def check_patterns(node, patterns):
    for pattern in patterns:
        found = node.find(string=pattern)
        if found:
            in_bp = sv.filter('.pg_boilerplate', found.parents)
            if not in_bp:
                return found

def mark_soup(soup):
    def mark_bp(node, mark, markers, top=True):
        divider = check_patterns(node, markers)
        if divider:
            # first, copy the element that contains the top (bottom) boilerplate divider 
            # and content that precedes (follows) it
            top_el = node
            if len(node.contents) == 1:
                # pathological
                top_el = node.contents[0]
            bp_section = soup.new_tag('section', id=mark)

            for elem in top_el.contents:
                bp_section.append(copy.copy(elem))
                if elem in divider.parents:
                    break

            # next remove the divider and anything before (after) the divider from the soup
            prune(node, divider, after=not top)
            divider.parent.extract()

            # remove anything after the divider from bp_header
            divider = check_patterns(bp_section, markers)
            # divider is a string
            prune(bp_section, divider, after=top)
            bp_section['class'] = 'pg_boilerplate'
            node.insert(0 if top else -1, bp_section)
            return True
        return False

    try:
        body = soup.html.body
    except:
        print('no body')
        return

    found_top = mark_bp(body, 'pg-header', TOP_MARKERS, top=True) 
    found_bottom = mark_bp(body, 'pg-footer', BOTTOM_MARKERS, top=False) 
    found_smallprint = mark_bp(body, 'pg-smallprint', SMALLPRINT_MARKERS, top=True)
    return found_top or found_bottom or found_smallprint


def strip_headers_from_txt(text):
    ''' 
    when input is plain text, strip the heaters and return (stripped_text, pg_header, pg_footer)
    '''
    def markers_split(text, markers):
        for marker in markers:
            divider = marker.search(text)
            if divider:
                (before, after) = text.split(divider.group(0))
                return before, divider.group(0), after
        return  None, text, None
    header_text, divider, text = markers_split(text, TOP_MARKERS)
    divider_tail = ''
    if '\n' in text:
        divider_tail, text = text.split('\n', maxsplit=1)
    pg_header = '\n'.join(['<pre id="pg-header">', header_text, divider, divider_tail, '</pre>'])

    text, divider, footer_text = markers_split(text, BOTTOM_MARKERS)
    pg_footer = '\n'.join(['<pre id="pg-footer">', divider, footer_text, '</pre>'])
    return text, pg_header, pg_footer



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

PicsDirWriter.py

Copyright 2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Copies pics into local directory. Needed for HTML and Xetex.

"""


import os.path

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.Logger import info, debug, error

from ebookmaker.parsers import webify_url
from ebookmaker import writers


class Writer(writers.BaseWriter):
    """ Writes Pics directory. """

    def copy_aux_files(self, job, dest_dir):
        """ Copy image files to dest_dir. Use image data cached in parsers. """

        for p in job.spider.parsers:
            if hasattr(p, 'resize_image') or hasattr(p, 'auxparser'):
                src_uri = p.attribs.url
                if src_uri.startswith(webify_url(dest_dir)):
                    debug('Not copying %s to %s: already there' % (src_uri, dest_dir))
                    continue

                fn_dest = gg.make_url_relative(webify_url(job.base_url), src_uri)
                fn_dest = os.path.join(dest_dir, fn_dest)

                # debug('base_url =  %s, src_uri = %s' % (job.base_url, src_uri))

                if gg.is_same_path(src_uri, fn_dest):
                    debug('Not copying %s to %s: same file' % (src_uri, fn_dest))
                    continue
                debug('Copying %s to %s' % (src_uri, fn_dest))

                fn_dest = gg.normalize_path(fn_dest)
                gg.mkdir_for_filename(fn_dest)
                try:
                    with open(fn_dest, 'wb') as fp_dest:
                        fp_dest.write(p.serialize())
                except IOError as what:
                    error('Cannot copy %s to %s: %s' % (src_uri, fn_dest, what))



    def build(self, job):
        """ Build Pics file. """

        dest_dir = os.path.abspath(job.outputdir)

        info("Creating Pics directory in: %s" % dest_dir)

        self.copy_aux_files(job, dest_dir)

        info("Done Pics directory in: %s" % dest_dir)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

Writer package

Copyright 2009-2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Base classes for *Writer modules. (EpubWriter, PluckerWriter, ...)

"""
import re
import subprocess

from functools import partial
import os.path

from lxml import etree
from lxml.builder import ElementMaker

from libgutenberg.Logger import debug, info, error
import libgutenberg.GutenbergGlobals as gg
from libgutenberg import MediaTypes

from ebookmaker import parsers
from ebookmaker import ParserFactory
from ebookmaker.CommonCode import Options
from ebookmaker.Version import VERSION, GENERATOR


options = Options()

def remove_cr(content):
    content = re.sub(r'\s*[\r\n]+\s*', '&#10;', content)
    return content

class BaseWriter(object):
    """
    Base class for EpubWriter, PluckerWriter, ...

    also used as /dev/null writer for debugging

    """

    VALIDATOR = None
    
    def build(self, job):
        """ override this in a real writer """
        pass


    @staticmethod
    def write_with_crlf(filename, bytes_):
        # \r\n is PG standard
        bytes_ = b'\r\n'.join(bytes_.splitlines()) + b'\r\n'

        # open binary so windows doesn't add another \r
        with open(filename, 'wb') as fp:
            fp.write(bytes_)


    def validate(self, job):
        """ Validate generated file using external tools. """

        if not self.VALIDATOR:
            return 0

        debug("Validating %s ..." % job.outputfile)

        filename = os.path.join(os.path.abspath(job.outputdir), job.outputfile)

        if hasattr(options.config, self.VALIDATOR):
            validator = getattr(options.config, self.VALIDATOR)
            info('validating...')
            params = validator.split() + [filename]
            checker = subprocess.Popen(params,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)

            (dummy_stdout, stderr) = checker.communicate()
            if stderr:
                error(stderr.decode("utf-8"))
                return 1

        info("%s validates ok." % job.outputfile)
        return 0


    def sync(self):
        """  Override this if you need to sync before program exit. """
        pass


    def make_links_relative(self, xhtml, base_url):
        """ Make absolute links in xhtml relative to base_url. """

        debug("Making links relative to: %s" % base_url)
        xhtml.rewrite_links(partial(gg.make_url_relative, base_url))



em = ElementMaker()

class HTMLishWriter(BaseWriter):
    """ Base class for writers with HTMLish contents. """

    @staticmethod
    def add_class(elem, class_):
        """ Add a class to html element. """

        classes = elem.get('class', '').split()
        classes.append(class_)
        elem.set('class', ' '.join(classes))


    @staticmethod
    def add_meta(xhtml, name, content):
        """ Add a meta tag. """

        for head in gg.xpath(xhtml, '//xhtml:head'):
            meta = em.meta(name=name, content=remove_cr(content))
            meta.tail = '\n'
            head.append(meta)

    @staticmethod
    def add_prop(xhtml, prop, content):
        """ Add a property meta tag. """

        for head in gg.xpath(xhtml, '//xhtml:head'):
            meta = em.meta(property=prop, content=remove_cr(content))
            meta.tail = '\n'
            head.append(meta)


    @staticmethod
    def add_meta_generator(xhtml):
        """ Add our piss mark. """

        HTMLishWriter.add_meta(xhtml, 'generator', GENERATOR % VERSION)


    @staticmethod
    def add_internal_css(xhtml, css_as_string):
        """ Add internal stylesheet to html. """

        if css_as_string and xhtml is not None:
            css_as_string = '\n' + css_as_string.strip(' \n') + '\n'
            for head in gg.xpath(xhtml, '//xhtml:head'):
                style = em.style(css_as_string, type='text/css')
                style.tail = '\n'
                head.append(style)

    @staticmethod
    def add_body_class(xhtml, classname):
        """ Add a class to the body element. """

        if classname and xhtml is not None:
            for body in gg.xpath(xhtml, '//xhtml:body'):
                HTMLishWriter.add_class(body, classname)


    def add_external_css(self, spider, xhtml, css_as_string, url):
        """ Add external stylesheet to html. """

        if css_as_string:
            attribs = parsers.ParserAttributes()
            attribs.orig_mediatype = attribs.HeaderElement('text/css')
            attribs.url = attribs.orig_url = url
            p = ParserFactory.ParserFactory.get(attribs)
            p.parse_string(css_as_string)
            p.make_links_absolute()
            spider.parsers.append(p)

        if xhtml is not None:
            for head in gg.xpath(xhtml, '//xhtml:head'):
                link = em.link(href=url, rel='stylesheet', type='text/css')
                link.tail = '\n'
                head.append(link)





# -*- coding: utf-8 -*-
# Copyright: This module is put into the public domain.
# Author: Marcello Perathoner <webmaster@gutenberg.org>

"""
Nroff writer for reStructuredText.

This module is more suitable for writing novel-type books than
documentation.

To process into plain text use:

  ``groff -t -K utf8 -T utf8 input.nroff > output.txt``

"""

from __future__ import unicode_literals

__docformat__ = 'reStructuredText'

import collections
import re

import six

from docutils import nodes, frontend
from docutils.writers.html4css1 import SimpleListChecker

from ebookmaker.mydocutils import writers
from ebookmaker.mydocutils.transforms import parts

from libgutenberg.Logger import info, debug, warning, error


LINE_WIDTH             = 66
TABLE_WIDTH            = 66 # max. table width

BLOCKQUOTE_INDENT      =  4
LIST_INDENT            =  2
FOOTNOTE_INDENT        =  6
CITATION_INDENT        = 10
FIELDLIST_INDENT       =  7
DEFINITION_LIST_INDENT =  7
OPTION_LIST_INDENT     =  7

NROFF_PREAMBLE = r""".\" -*- mode: nroff -*- coding: {encoding} -*-
.\" This file produces plain text. Usage:
.\"   $ groff -t -K {device} -T {device} this_file > output.txt
.
.ll 72m
.po 0
.ad l           \" text-align: left
.nh             \" hyphenation: off
.cflags 0 .?!   \" single sentence space
.cflags 0 -\[hy]\[em]   \" don't break on -
.
.de nop
..
.blm nop        \" do nothing on empty line
.
.nr [env_cnt] 0
.ev 0           \" start in a defined environment
.
.de push_env
.br
.nr last_env \\n[.ev]            \" save current environment name
.nr env_cnt +1   \" generate new environment name
.ev \\n[env_cnt]
.evc \\n[last_env]
..
.de pop_env
.br
.ev
.nr env_cnt -1
..
.
"""

NROFF_POSTAMBLE = r""".
.\" End of File
"""

def nroff_units (length_str, reference_length = None):
    """ Convert rst units to Nroff units. """

    match = re.match (r'(\d*\.?\d*)\s*(\S*)', length_str)
    if not match:
        return length_str

    value, unit = match.groups ()

    if unit in ('', 'u'):
        return value

    # percentage: relate to current line width
    elif unit == '%':
        reference_length = reference_length or LINE_WIDTH
        return int (float (value) / 100.0 * reference_length)

    return length_str


class InheritTransform (parts.InheritTransform):
    pass


class Writer (writers.Writer):
    """ A plaintext writer thru nroff. """

    supported = ('nroff',)
    """Formats this writer supports."""

    output = None
    """Final translated form of `document`."""

    settings_spec = (
        'Nroff-Specific Options',
        None,
        (('Should lists be compacted when possible?',
          ['--compact-lists'],
          {'default': 1,
           'action': 'store_true',
           'validator': frontend.validate_boolean}),
         ('Which encoding are we targeting? (As hint to the writer.).',
          ['--encoding'],
          {'default': 'utf-8',
           'validator': frontend.validate_encoding}),
         ))

    settings_defaults = {
        'encoding': 'utf-8',
        }

    config_section = 'NROFF writer'

    def get_transforms (self):
        tfs = writers.Writer.get_transforms (self)
        return tfs + [parts.TextNodeWrapper, InheritTransform]

    def __init__ (self):
        writers.Writer.__init__ (self)
        self.translator_class = Translator


class TablePass2 (nodes.SparseNodeVisitor):

    """
    Makes a second pass over table to build tbl format specification.
    """

    def __init__ (self, document, table, rows, cols):
        nodes.SparseNodeVisitor.__init__ (self, document)
        self.cols = cols
        self.types = ['-'] * (rows * cols)
        self.i = 0

        self.table_width = nroff_units (table.get ('width', '100%'), TABLE_WIDTH)
        colspecs = list (table.traverse (nodes.colspec))
        self.table_width -= len (colspecs) * 2

    def visit_entry (self, node):
        """ Called on each table cell. """

        if 'vspan' in node:
            raise nodes.SkipNode

        rows = node.get ('morerows', 0) + 1
        cols = node.get ('morecols', 0) + 1

        for j in range (0, cols):
            self.types[self.i + j] = 's'
            for k in range (1, rows):
                self.types[self.i + (k * self.cols) + j] = '^'

        align = node.colspecs[0].get ('align', 'left')
        align = { 'right': 'r', 'center': 'c' }.get (align, 'l')
        self.types[self.i] = align # l, r or c

        valign = node.colspecs[0].get ('valign', 'middle')
        valign = { 'top': 't', 'bottom': 'd' }.get (valign, '')
        self.types[self.i] += valign # t or d

        if len (node.colspecs) == 1: # no span
            self.types[self.i] += 'w(%d)' % (
                node.colspecs[0]['relative_width'] * self.table_width)

        while self.i < len (self.types) and self.types[self.i] != '-':
            self.i += 1

        raise nodes.SkipNode

    def build_format_spec (self):
        """ Build a tbl format specification for this table. """
        types = zip (*[iter (self.types)] * self.cols) # cluster by row

        types = [' '.join (x) for x in types]
        types = ','.join (types)

        # print types

        return '%s.\n' % types


class Translator (writers.Translator):
    """ nroff translator """

    superscript_digits = '⁰¹²³⁴⁵⁶⁷⁸⁹'

    subscript_digits   = '₀₁₂₃₄₅₆₇₈₉'

    def __init__ (self, document):
        writers.Translator.__init__ (self, document)
        self.settings = document.settings

        self.encoding = self.settings.encoding
        # the name of the groff device
        self.device = { 'utf-8':      'utf8',
                        'iso-8859-1': 'latin1',
                        'us-ascii':   'ascii' }.get (self.encoding, '<device>')
        self.init_translate_maps ()
        self.body = []
        self.context = self.body # start with context == body
        self.docinfo = collections.defaultdict (list)
        self.list_enumerator_stack = []
        self.section_level = 0
        self.vspace = 0 # pending space (need this for collapsing)
        self.last_output_char = '\n' # used to make sure we output \n before commands
        self.field_name = None
        self.compacting = 0 # > 0 if we are inside a compacting list
        self.in_literal = 0 # are we inside one or more literal blocks?


    def preamble (self):
        """ Inserts nroff preamble. """
        return NROFF_PREAMBLE.format (encoding = self.encoding, device = self.device)


    def postamble (self):
        """ Inserts nroff postamble. """
        return NROFF_POSTAMBLE.format (encoding = self.encoding, device = self.device)


    def init_translate_maps (self):
        # see: man groff_char
        self.translate_map = {
            0x0027: r"\(cq",     # ' is non-breaking command
            0x002e: r"\N'46'",   # . is breaking command
            0x005c: r"\(rs",     # \ is escape char
            0x00a0: r"\ ",       # nbsp
            0x00ad: r"\%",       # shy
            0x03c6: r"\(+F",     # curly (text) phi,  | groff inverts
            0x03d5: r"\(*F",     # stroked (math) phi | these two
            }

        self.translate_map_literal = {}
        self.translate_map_literal.update (self.translate_map)
        self.translate_map_literal.update ({
                # by default groff substitutes these chars with chars
                # better suited for typography, but less suited for
                # source code display.  we want the original chars.
                # see: man groff_char
                0x0027: r'\(aq', # apos
                0x002d: r'\-',   # hyphen-minus
                0x005e: r'\(ha', # circumflex
                0x0060: r'\`',   # backtick
                0x007e: r'\(ti', # tilde
                })


    def translate (self, text):
        """ Reduce the charset while keeping text a unicode string. """

        if self.in_literal:
            text = text.translate (self.translate_map_literal)
        else:
            text = text.translate (self.translate_map)

        return text


    def register_classes (self):
        """ Register classes.

        Use a fairly general set of font attributes.

        """

        self.register_class ('simple', 'left',         '.ad l',   '')
        self.register_class ('simple', 'center',       '.ad c',   '')
        self.register_class ('simple', 'centerleft',   '.ad c',   '')
        self.register_class ('simple', 'right',        '.ad r',   '')

        self.register_class ('inline', 'italics',      r'\fI',    r'\fP')
        self.register_class ('inline', 'bold',         r'\fB',    r'\fP')
        self.register_class ('inline', 'monospaced',   r'\fM',    r'\fP')
        self.register_class ('inline', 'superscript',  r'\s-2\u', r'\d\s0')
        self.register_class ('inline', 'subscript',    r'\s-2\d', r'\u\s0')
        self.register_class ('inline', 'small-caps',   r'\fI',    r'\fP')
        self.register_class ('inline', 'gesperrt',     r'\fI',    r'\fP')
        self.register_class ('inline', 'antiqua',      r'\fI',    r'\fP')
        self.register_class ('inline', 'larger',       r'\fI',    r'\fP')
        self.register_class ('inline', 'smaller',      r'\fI',    r'\fP')


    def set_class_on_child (self, node, class_, index = 0):
        """
        Set class `class_` on the visible child no. index of `node`.
        Do nothing if node has fewer children than `index`.
        """
        children = [n for n in node if not isinstance (n, nodes.Invisible)]
        try:
            child = children[index]
        except IndexError:
            return
        child['classes'].append (class_)

    def set_first_last (self, node):
        """ Set class 'first' on first child, 'last' on last child. """
        self.set_class_on_child (node, 'first', 0)
        self.set_class_on_child (node, 'last', -1)

    def cmd (self, cmds):
        """ Add nroff commands. """
        if isinstance (cmds, six.string_types):
            cmds = [cmds]

        if self.last_output_char != '\n':
            self.context.append ('\n')
        for c in cmds:
            if c:
                self.context.append (".%s\n" % c)
                self.last_output_char = '\n'

    def text (self, text):
        """ Output text. """
        if isinstance (text, six.string_types):
            text = [text]

        for t in text:
            if t:
                self.context.append (t)
                self.last_output_char = t[-1]

    def text_or_cmd (self, text_or_cmds):
        """ Try to output the string appropriately. """
        if isinstance (text_or_cmds, six.string_types):
            text_or_cmds = [text_or_cmds]

        for item in text_or_cmds:
            if not item:
                continue
            if item[0] == '.':
                self.cmd (item[1:])
            else:
                self.text (item)

    def backspace (self):
        """ Remove last character from output. """
        self.context[-1] = self.context[-1][:-1]
        while not self.context[-1]:
            self.context.pop ()
        if self.context and self.context[-1]:
            self.last_output_char = self.context[-1][-1]

    def comment (self, text):
        """ Output nroff comment. """
        self.cmd ('\\"%s' % self.translate (text))

    def output_sp (self):
        """ Output spacing and pending stuff. """
        if self.vspace == 1999: # magic number to eat all space
            self.vspace = 0
        if self.vspace:
            self.cmd ('sp' if self.vspace == 1 else 'sp %d' % self.vspace)
            self.vspace = 0

    def br (self):
        """ Insert br command. """
        if self.context[-1] == '.br\n':
            self.cmd ('sp')
        self.cmd ('br')

    def ta (self, indent, text):
        """ Right-tabulate text to indent position. """
        if indent > 1:
            self.cmd ('ta %dmR' % (indent - 1))
            self.text ('\t')
        self.text (text)

    def push (self):
        """ Push environment. """
        self.cmd ('push_env')

    def pop (self):
        """ Pop environment. """
        self.cmd ('pop_env')

    def as_superscript (self, n):
        """ Return n as string using superscript unicode chars. """
        if self.encoding != 'utf-8':
            return '[%d]' % n
        res = ''
        for d in str (n):
            res += self.superscript_digits [int (d)]
        return res

    def br_if_line_longer_than (self, length):
        """ Insert a line break if the line was longer than length.

        Use this to compact lists etc. """

        # we have to break the line to get the length,
        # then we eventually go back
        self.cmd (('br', 'if (\\n[.n] < %dm) .sp -1' % length))
        self.sp (0)

    def indent (self, by = 2):
        """ Indent text. """
        self.cmd ('in +%dm' % by)

    def rindent (self, by = 2):
        """ Indent text on the right side. """
        self.cmd ('ll -%dm' % by)


    def visit_outer (self, node):
        """ Called before visiting a node. """
        # self.comment ("visiting outer: %s" % node.__class__.__name__)

        if node.type != 'text' and 'margin' in node['classes']:
            self.sp ()

        if node.type in ('compound', 'simple'):
            if not isinstance (node, nodes.entry):
                # tbl preprocessor eats some contents of rows
                self.push ()

    def visit_inner (self, node):
        """ Called after visiting a node. """

        # self.comment ("visiting inner: %s" % node.__class__.__name__)

        if node.type == 'simple':
            self.output_sp ()

        if node.type != 'text':
            self.text_or_cmd (self.get_prefix (node.type, node['classes']))

    def depart_inner (self, node):
        """ Called before departing from a node. """

        if node.type != 'text':
            self.text_or_cmd (self.get_suffix (node.type, node['classes']))

        # self.comment ("departing inner: %s" % node.__class__.__name__)

    def depart_outer (self, node):
        """ Called after departing from a node. """

        if node.type != 'text' and 'margin' in node['classes']:
            self.sp ()

        if node.type in ('compound', 'simple'):
            if not isinstance (node, nodes.entry):
                self.pop ()

        # self.comment ("departing outer: %s" % node.__class__.__name__)


    def visit_Text (self, node):
        text = self.translate (node.astext ())

        if hasattr (node, 'attributes') and 'white-space-pre-line' in node.attributes['classes']:
            for line in text.splitlines (True):
                self.text (line)
                if '\n' in line:
                    self.br ()
        else:
            self.text (text)


    def depart_Text (self, node):
        pass


    def visit_inline (self, node):
        pass

    def depart_inline (self, node):
        pass


    # start docinfo elements (parse only for now)

    def visit_docinfo (self, node):
        pass

    def depart_docinfo (self, node):
        pass

    def visit_authors (self, node):
        pass

    def depart_authors (self, node):
        pass

    def visit_field (self, node):
        pass

    def depart_field (self, node):
        pass

    def visit_field_name (self, node):
        self.field_name = node.astext ().lower ().replace (' ', '_')
        raise nodes.SkipNode

    def depart_field_name (self, node):
        pass

    def visit_field_body (self, node, name = None):
        # name either from element or stored by <field_name>
        self.context = self.docinfo[name or self.field_name]

    def depart_field_body (self, node):
        self.context = self.body

    def visit_field_list (self, node):
        self.sp ()
        self.indent (FIELDLIST_INDENT)

    def depart_field_list (self, node):
        pass

    # start admonitions

    def visit_admonition (self, node, name = None):
        if name:
            self.text (name)
            self.sp ()
        else:
            self.br ()
        self.indent (BLOCKQUOTE_INDENT)

    def depart_admonition (self, node):
        pass

    # start definition lists

    def visit_definition_list (self, node):
        self.sp ()
        # self.output_sp ()

    def depart_definition_list (self, node):
        pass

    def visit_definition_list_item (self, node):
        pass

    def depart_definition_list_item (self, node):
        pass

    def visit_term (self, node):
        pass

    def depart_term (self, node):
        self.br_if_line_longer_than (DEFINITION_LIST_INDENT)

    def visit_classifier (self, node):
        pass

    def depart_classifier (self, node):
        pass

    def visit_definition (self, node):
        self.indent (DEFINITION_LIST_INDENT)

    def depart_definition (self, node):
        self.sp ()

    # start option lists

    def visit_option_list (self, node):
        self.sp ()
        # self.output_sp ()

    def depart_option_list (self, node):
        self.output_sp () # clear sp (0)
        self.sp ()

    def visit_option_list_item (self, node):
        pass

    def depart_option_list_item (self, node):
        pass

    def visit_option_group (self, node):
        pass

    def depart_option_group (self, node):
        self.br_if_line_longer_than (OPTION_LIST_INDENT)

    def visit_option (self, node):
        pass

    def depart_option (self, node):
        if 'last' not in node['classes']:
            self.text (', ')

    def visit_option_string (self, node):
        pass

    def depart_option_string (self, node):
        pass

    def visit_option_argument (self, node):
        self.text (node.get ('delimiter', ' '))
        self.text (r'\fI')

    def depart_option_argument (self, node):
        self.text (r'\fP')

    def visit_description (self, node):
        self.indent (OPTION_LIST_INDENT)

    def depart_description (self, node):
        self.sp (0)

    # lists

    def check_simple_list (self, node):
        """Check for a simple list that can be rendered compactly."""
        try:
            node.walk (SimpleListChecker (self.document))
            return True
        except nodes.NodeFound:
            return False

    def is_compactable (self, node):
        return ('compact' in node['classes']
                or (self.settings.compact_lists
                    and 'open' not in node['classes']
                    and (# self.compact_simple or
                         # self.topic_classes == ['contents'] or
                         self.check_simple_list (node))))

    def list_start (self, node):
        if not isinstance (node.parent, nodes.list_item):
            self.sp ()
            self.output_sp () # list entry will eat all space, so output it now
            self.indent (LIST_INDENT)
        self.compacting += self.is_compactable (node)
        self.comment ('compacting: %d' % self.compacting)
        self.list_enumerator_stack.append (writers.ListEnumerator (node, self.encoding))

    def list_end (self, node):
        self.list_enumerator_stack.pop ()
        self.compacting = max (0, self.compacting - 1)
        if not isinstance (node.parent, nodes.list_item):
            self.output_sp ()
            self.sp ()

    def visit_list_item (self, node):
        self.sp (0)
        self.br ()
        indent = self.list_enumerator_stack[-1].get_width ()
        label = self.list_enumerator_stack[-1].get_next ()
        if label:
            self.ta (indent, label)
            self.br_if_line_longer_than (indent)
        self.push ()
        self.indent (indent)

    def depart_list_item (self, node):
        self.pop ()
        if self.compacting:
            self.sp (0)
        else:
            self.sp ()
            self.output_sp ()

    def visit_bullet_list (self, node):
        self.list_start (node)

    def depart_bullet_list (self, node):
        self.list_end (node)

    def visit_enumerated_list (self, node):
        self.list_start (node)

    def depart_enumerated_list (self, node):
        self.list_end (node)

    # end lists

    def visit_block_quote (self, node):
        self.set_first_last (node)
        self.sp ()
        self.indent (BLOCKQUOTE_INDENT)
        self.rindent (BLOCKQUOTE_INDENT)

    def depart_block_quote (self, node):
        classes = node['classes']
        if 'epigraph' in classes:
            self.sp (2)
        if 'highlights' in classes:
            self.sp (2)

    def visit_comment (self, node):
        for line in node.astext ().splitlines ():
            self.comment (line)
        raise nodes.SkipNode

    def visit_container (self, node):
        pass

    def depart_container (self, node):
        pass

    def visit_compound (self, node):
        pass

    def depart_compound (self, node):
        pass

    def visit_decoration (self, node):
        pass

    def depart_decoration (self, node):
        pass

    def visit_doctest_block (self, node):
        self.visit_literal_block (node)

    def depart_doctest_block (self, node):
        self.depart_literal_block (node)

    def visit_document (self, node):
        pass

    def depart_document (self, node):
        pass

    def visit_footer (self, node):
        self.document.reporter.warning (
            'footer not supported', base_node = node)

    def depart_footer (self, node):
        pass

    # footnotes, citations, labels

    def visit_label (self, node):
        # footnote and citation
        indent = 0
        if isinstance (node.parent, nodes.footnote):
            indent = FOOTNOTE_INDENT
        elif isinstance (node.parent, nodes.citation):
            indent = CITATION_INDENT
        else:
            self.document.reporter.warning ('label unsupported',
                base_node = node)

        try:
            label = self.as_superscript (int (node.astext ()))
        except ValueError:
            label = '[%s]' % self.translate (node.astext ())

        if label:
            self.ta (indent, label)
            self.br_if_line_longer_than (indent)
        self.push ()
        self.indent (indent)
        raise nodes.SkipNode

    def depart_label (self, node):
        pass

    def visit_footnote (self, node):
        self.sp ()
        self.output_sp ()

    def depart_footnote (self, node):
        self.pop ()
        self.sp ()

    def visit_footnote_reference (self, node):
        try:
            self.text (
                self.as_superscript (int (node.astext ())))
        except ValueError:
            self.text (node.astext ())
        raise nodes.SkipNode

    def visit_citation (self, node):
        self.visit_footnote (node)

    def depart_citation (self, node):
        self.depart_footnote (node)

    def visit_citation_reference (self, node):
        self.text ('[%s]' % self.translate (node.astext ()))
        raise nodes.SkipNode

    # end footnotes

    def visit_generated (self, node):
        pass

    def depart_generated (self, node):
        pass

    def visit_header (self, node):
        self.document.reporter.warning (
            'header not supported', base_node = node)

    def depart_header (self, node):
        pass

    def visit_figure (self, node):
        self.sp (2)
        self.indent (BLOCKQUOTE_INDENT)
        self.rindent (BLOCKQUOTE_INDENT)

    def depart_figure (self, node):
        self.sp (2)

    def visit_image (self, node):
        self.text ('%s' % self.translate (node.attributes.get ('alt', '[image]')))

    def depart_image (self, node):
        pass

    def visit_caption (self, node):
        pass

    def depart_caption (self, node):
        pass

    def visit_legend (self, node):
        pass

    def depart_legend (self, node):
        pass

    def visit_line_block (self, node):
        if not isinstance (node.parent, nodes.line_block):
            # outermost block
            self.sp ()
        else:
            self.br ()
        if isinstance (node.parent, nodes.line_block):
            self.indent ()
        self.cmd ('ad l')

    def depart_line_block (self, node):
        if not isinstance (node.parent, nodes.line_block):
            # outermost block
            self.sp ()

    def visit_line (self, node):
        if node.children:
            self.push ()
            self.indent (8)
            self.cmd ('ti -8')
        else:
            # empty line
            self.sp ()
            self.output_sp ()

    def depart_line (self, node):
        if node.children:
            self.br ()
            self.pop ()

    def visit_literal_block (self, node):
        self.sp ()
        self.indent (BLOCKQUOTE_INDENT)
        self.cmd (('nf', 'ft C', 'blm'))
        self.in_literal += 1

    def depart_literal_block (self, node):
        self.in_literal -= 1
        self.cmd ('blm nop')
        self.sp ()

    #
    #
    #

    def visit_paragraph (self, node):
        self.sp ()
        # self.output_sp ()

    def depart_paragraph (self, node):
        self.sp ()

    def visit_section (self, node):
        self.section_level += 1

    def depart_section (self, node):
        self.section_level -= 1

    def visit_raw (self, node):
        if 'nroff' in node.get ('format', '').split():
            raw = node.astext ()
            if raw[0] == '.':
                self.cmd (raw[1:])
            else:
                self.text (raw)

        # ignore other raw formats
        raise nodes.SkipNode

    def visit_substitution_definition (self, node):
        """Internal only."""
        raise nodes.SkipNode

    def visit_substitution_reference (self, node):
        self.document.reporter.warning ('"substitution_reference" not supported',
                base_node=node)

    def visit_target (self, node):
        # internal hyperlink target, no such thing in nroff
        pass

    def depart_target (self, node):
        pass

    def visit_system_message (self, node):
        self.sp ()
        line = ', line %s' % node['line'] if 'line' in node else ''
        self.text ('"System Message: %s/%s (%s:%s)"'
                  % (node['type'], node['level'], node['source'], line))
        self.sp ()
        self.indent (BLOCKQUOTE_INDENT)

    def depart_system_message (self, node):
        self.sp ()

    # tables

    def visit_table (self, node):
        self.sp (2)
        pass_1 = writers.TablePass1 (self.document)
        node.walk (pass_1)
        rows = pass_1.rows ()
        cols = pass_1.cols ()

        pass_2 = TablePass2 (self.document, node, rows, cols)
        node.walk (pass_2)
        node.pass_2 = pass_2

    def depart_table (self, node):
        self.cmd ('TE')
        self.sp (2)

    def visit_table_title (self, node):
        pass

    def depart_table_title (self, node):
        self.sp ()

    def visit_tgroup (self, node):
        table = node.parent
        # output this after table caption
        self.output_sp ()
        self.cmd ('TS')
        options = ['center']
        self.text (' '.join (options) + ';\n')
        self.text (node.parent.pass_2.build_format_spec ())

    def depart_tgroup (self, node):
        pass

    def visit_colspec (self, node):
        pass

    def depart_colspec (self, node):
        pass

    def visit_thead (self, node):
        self.set_first_last (node) # mark first row of head
        self.text ('=\n')

    def depart_thead (self, node):
        pass

    def visit_tbody (self, node):
        self.text ('=\n')
        self.set_first_last (node) # mark first row of body

    def depart_tbody (self, node):
        self.text ('=')

    def visit_row (self, node):
        self.set_first_last (node) # mark first and last cell
        if 'first' not in node['classes']:
            if 'rows' in node.parent.parent.parent['hrules']:
                self.text ('_\n')
            else:
                self.sp ()

    def depart_row (self, node):
        pass

    def visit_entry (self, node):
        self.sp (0)
        if 'first' in node['classes']:  # first cell in row
            self.text ('T{\n')
            self.last_output_char = '\n'

    def depart_entry (self, node):
        # self.sp (0)
        if self.last_output_char != '\n':
            self.context.append ('\n')
        if 'last' in node['classes']:  # last cell in row
            self.text ('T}\n')
        else:
            self.text ('T}\tT{\n')
        self.last_output_char = '\n'

    # end tables

    def visit_document_title (self, node):
        self.cmd ('ad c')

    def depart_document_title (self, node):
        if 'with-subtitle' in node['classes']:
            self.sp (1)
        else:
            self.sp (2)

    def visit_document_subtitle (self, node):
        self.sp (1)
        self.cmd ('ad c')

    def depart_document_subtitle (self, node):
        self.sp (2)

    def visit_section_title (self, node):
        self.sp (3)

    def depart_section_title (self, node):
        self.sp (2)

    def visit_section_subtitle (self, node):
        self.sp (1)

    def depart_section_subtitle (self, node):
        self.sp (2)

    def visit_topic (self, node):
        self.sp (4)

    def depart_topic (self, node):
        self.sp (4)

    def visit_topic_title (self, node):
        pass

    def depart_topic_title (self, node):
        self.sp (2)

    def visit_sidebar (self, node):
        pass

    def depart_sidebar (self, node):
        pass

    def visit_rubric (self, node):
        pass

    def depart_rubric (self, node):
        pass

    def visit_page (self, node):
        if 'vspace' in node['classes']:
            self.sp (node['length'])
        elif 'clearpage' in node['classes']:
            self.cmd ('bp')
        elif 'cleardoublepage' in node['classes']:
            self.cmd ('bp')
        elif 'vfill' in node['classes']:
            self.sp (4)

    def depart_page (self, node):
        pass

    def visit_newline (self, node):
        self.br ()

    def depart_newline (self, node):
        pass


    def visit_transition (self, node):
        pass

    def depart_transition (self, node):
        pass


    def visit_attribution (self, node):
        pass

    def depart_attribution (self, node):
        pass


    def visit_problematic (self, node):
        self.cmd ('nf')

    def depart_problematic (self, node):
        self.cmd ('fi')

    def visit_meta (self, node):
        raise NotImplementedError (node.astext ())

    def unimplemented_visit (self, node):
        raise NotImplementedError ('visiting unimplemented node type: %s'
                                  % node.__class__.__name__)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
PushPackager.py

Copyright 2011 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Package a zip containing everything, that can be pushed to ibiblio.

"""


import os.path
import re

from libgutenberg.Logger import info, error
import libgutenberg.GutenbergGlobals as gg

from ebookmaker.CommonCode import Options
from ebookmaker.packagers import ZipPackager

options = Options()
TYPE = 'ww'
FORMATS = ['push']

class Packager (ZipPackager):
    """ Package one big zip for push.

    Zip contains one directory named after ebook_no.
    This dir mirrors structure on ibiblio::

      12345/12345.txt
      12345/12345.zip
      12345/12345-h/12345-h.html
      12345/12345-h/images/cover.jpg
      12345/12345-h.zip

    """

    def package (self, job):
        self.setup (job)
        zipfilename = job.outputfile # filename is zipfile

        m = re.match (r'\d+', zipfilename)
        if m:
            ebook_no = m.group (0)
        else:
            error ('Invalid filename %s for push packager.' % zipfilename)
            return

        zip_ = self.create (zipfilename)

        for suffix in '.txt -8.txt -0.txt .zip -8.zip -0.zip -rst.zip -h.zip'.split ():
            filename = '%s%s' % (ebook_no, suffix)
            memberfilename = '%s/%s' % (ebook_no, filename)
            self.add (zip_, filename, memberfilename)

        for suffix, ext in (('-h', 'html'), ('-rst', 'rst')):
            filename = '%s%s.%s' % (ebook_no, suffix, ext)
            memberfilename = '%s/%s%s/%s' % (ebook_no, ebook_no, suffix, filename)
            self.add (zip_, filename, memberfilename)

            # image files
            for url in options.html_images_list:
                rel_url = gg.make_url_relative (job.base_url, url)
                filename = os.path.join (self.path, rel_url)
                memberfilename = '%s/%s%s/%s' % (ebook_no, ebook_no, suffix, rel_url)
                self.add (zip_, filename, memberfilename)

        zip_.close ()
        info ('Done Zip file: %s' % zipfilename)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

Spider.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Rudimentary Web Spider

"""

import os.path
import fnmatch

from six.moves import urllib

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import NS
from libgutenberg.Logger import critical, debug, error, info, warning
from libgutenberg import MediaTypes

from ebookmaker import parsers
from ebookmaker.CommonCode import Options
from ebookmaker.ParserFactory import ParserFactory

options = Options()

class Spider(object):
    """ A very rudimentary web spider. """

    def __init__(self, job):
        self.parsed_urls = set()
        self.parsers = []
        self.redirection_map = {}

        dirpath = os.path.dirname(job.url)  # platform native path
        # use for parser only
        self.include_urls = []
        self.include_urls += options.include_urls or [parsers.webify_url(dirpath) + '/*']

        self.include_mediatypes = []
        self.include_mediatypes += options.include_mediatypes
        if job.subtype == '.images' or job.type == 'rst.gen':
            self.include_mediatypes.append('image/*')
        if job.type == 'html.images':
            self.include_mediatypes.append('*/*')
        self.exclude_urls = []
        self.exclude_urls += options.exclude_urls

        self.exclude_mediatypes = []
        self.exclude_mediatypes += options.exclude_mediatypes

        self.max_depth = options.max_depth or six.MAXSIZE
        self.jobtype = job.type
        self.job_dc = job.dc


    def recursive_parse(self, root_attribs):
        """ Do a recursive parse starting from url.

        Do a breadth-first traversal. Assuming the first page contains
        a linked TOC, this will get us a more natural ordering of the
        pages than a depth-first traversal.

        """

        queue = []

        debug("Start of retrieval")

        # enqueue root url

        counter = 0
        self.enqueue(queue, 0, root_attribs, True)

        while queue:
            depth, attribs = queue.pop(0)

            url = self.redirect(attribs.url)
            if url in self.parsed_urls:
                continue

            parser = ParserFactory.create(url, attribs)
            if parser is None:
                continue
            # Maybe the url was redirected to something we already have?
            url = parser.attribs.url
            if url in self.parsed_urls:
                critical('no content in %s', url)
                continue
            self.parsed_urls.add(url)
            if hasattr(parser, 'add_title'):
                parser.add_title(self.job_dc)

            self.add_redirection(parser.attribs.orig_url, url)
            parser.pre_parse()
            self.parsers.append(parser)

            # look for more documents to add to the queue
            debug("Requesting iterlinks for: %s ..." % url)
            for url, elem in parser.iterlinks():
                counter += 1
                url = urllib.parse.urldefrag(url)[0]
                if url == parser.attribs.url or url in self.parsed_urls:
                    continue
                if elem.get('rel') == 'nofollow' and self.jobtype in ('epub.images', 'epub3.images'):
                    # remove link to content not followed
                    elem.tag = 'span'
                    elem.set('data-nofollow-href', elem.get('href'))
                    del elem.attrib['href']
                    del elem.attrib['rel']
                    warning('not followed: %s' % url)
                    continue

                new_attribs = parsers.ParserAttributes()
                new_attribs.url = url
                new_attribs.referrer = parser.attribs.url

                for k, v in elem.items():
                    if k in ('id', 'title'):
                        setattr(new_attribs, k, v)
                    elif k == 'type':
                        new_attribs.orig_mediatype = new_attribs.HeaderElement.from_str(v)
                    elif k == 'rel':
                        new_attribs.rel.update(v.lower().split())

                if not new_attribs.id:
                    # synthesize and set an id for backlink
                    seed = url + ' ' + str(counter)
                    new_attribs.id = f'id-{abs(hash(seed))}'
                    elem.attrib['id'] = new_attribs.id
   
                tag = elem.tag
                if tag == NS.xhtml.a:
                    if self.is_image(new_attribs) and self.is_included_url(new_attribs) and \
                            self.is_included_mediatype(new_attribs) and \
                            self.jobtype in ('epub.images', 'epub3.images'):
                        # need to wrap an image
                        wrapper_parser = parsers.WrapperParser.Parser(new_attribs)
                        if wrapper_parser.attribs.url not in self.parsed_urls:
                            ParserFactory.parsers[wrapper_parser.attribs.url] = wrapper_parser
                            self.parsers.append(wrapper_parser)
                            self.parsed_urls.add(wrapper_parser.attribs.url)
                        
                        # careful! we're changing the xml here, and the parser gets cached,
                        # so make sure 'epub.images' is the last job done.
                        elem.set('href', wrapper_parser.attribs.url)
                        new_attribs.referrer = wrapper_parser.attribs.url
                        elem.set('title', wrapper_parser.attribs.title)
                        self.enqueue(queue, depth + 1, new_attribs, False)
                    else:
                        self.enqueue(queue, depth + 1, new_attribs, True)
                        
                elif tag in (NS.xhtml.img, NS.xhtml.style):
                    self.enqueue(queue, depth, new_attribs, False)
                elif tag == NS.xhtml.link:
                    if new_attribs.rel.intersection(('stylesheet', 'icon')):
                        self.enqueue(queue, depth, new_attribs, False)
                    else:
                        self.enqueue(queue, depth + 1, new_attribs, True)
                elif tag == NS.xhtml.object:
                    self.enqueue(queue, depth, new_attribs, False)

        debug("End of retrieval")

        # rewrite redirected urls
        if self.redirection_map:
            for parser in self.parsers:
                parser.remap_links(self.redirection_map)
        # remove parsers with missing content
        self.parsers = [parser for parser in self.parsers if parser.fp != None]

        self.topological_sort()


    def enqueue(self, queue, depth, attribs, is_doc):
        """ Enqueue url for parsing."""
        if is_doc:
            if not self.is_included_url(attribs):
                warning('External link in %s: %s' % (attribs.referrer, attribs.url))
                return
            if depth >= self.max_depth:
                critical('Omitted file %s due to depth > max_depth' % attribs.url)
                return
        if not self.is_included_mediatype(attribs) and not self.is_included_relation(attribs):
            return
        elif not self.is_included_url(attribs) and not self.is_included_relation(attribs):
            critical('Failed for embedded media in %s from disallowed location: %s'
                  % (attribs.referrer, attribs.url))
            return

        queue.append((depth, attribs))


    def is_image(self, attribs):
        """ Return True if png, gif, svg or jpg. """
        return self.get_mediatype(attribs) in parsers.ImageParser.mediatypes


    def is_included_url(self, attribs):
        """ Return True if this document is eligible. """

        url = attribs.url

        included = any([fnmatch.fnmatchcase(url, x) for x in self.include_urls])
        excluded = any([fnmatch.fnmatchcase(url, x) for x in self.exclude_urls])

        if included and not excluded:
            return True

        if excluded:
            debug("Dropping excluded %s" % url)
        if not included:
            debug("Dropping not included %s" % url)
        return False


    def get_mediatype(self, attribs):
        """ Get mediatype out of attribs, guessing if needed. """
        if attribs.orig_mediatype is None:
            mediatype = MediaTypes.guess_type(attribs.url)
            if mediatype:
                attribs.orig_mediatype = attribs.HeaderElement(mediatype)
            else:
                return None
        return attribs.orig_mediatype.value


    def is_included_mediatype(self, attribs):
        """ Return True if this document is eligible. """

        mediatype = self.get_mediatype(attribs)
        if not mediatype:
            warning('Mediatype could not be determined from url %s' % attribs.url)
            return True # always include if mediatype unknown

        included = any([fnmatch.fnmatch(mediatype, pattern)
                        for pattern in self.include_mediatypes])
        excluded = any([fnmatch.fnmatch(mediatype, pattern)
                        for pattern in self.exclude_mediatypes])

        if included and not excluded:
            return True

        if excluded:
            debug("Dropping excluded mediatype %s" % mediatype)
        if not included:
            debug("Dropping not included mediatype %s" % mediatype)

        return False


    def is_included_relation(self, attribs):
        """ Return True if this document is eligible. """

        keep = attribs.rel.intersection(('icon', 'important', 'linked_image'))
        if keep:
            debug("Not dropping after all because of rel.")

        return keep


    def topological_sort(self):
        """ Do a topological sort of documents using <link rel='next'> """

        relnext = [(p.attribs.referrer, p.attribs.url)
                   for p in self.parsers if 'next' in p.attribs.rel]
        if relnext:
            try:
                d = {}
                for order, url in enumerate(gg.topological_sort(relnext)):
                    d[url] = order
                    debug("%s order %d" % (url, order))
                for parser in self.parsers:
                    parser.order = d.get(parser.attribs.url, 999999)
                self.parsers.sort(key=lambda p: p.order)

            except Exception:
                pass


    def add_redirection(self, from_url, to_url):
        """ Remember this redirection. """

        if from_url != to_url:
            self.redirection_map[from_url] = to_url
            debug("Adding redirection from %s to %s" % (from_url, to_url))


    def redirect(self, url):
        """
        Redirect url.

        Parsers are cached under the 200 url. This is an offline redirect
        to find the 200 url.

        """
        return self.redirection_map.get(url, url)


    def dict_urls_mediatypes(self):
        """ Return a dict of all parsed urls and mediatypes. """
        return dict([(p.attribs.url, p.mediatype()) for p in self.parsers])


    def aux_file_iter(self):
        """ Iterate over image files. Return absolute urls. """

        for p in self.parsers:
            if hasattr(p, 'resize_image'):
                yield p.attribs.url



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
PDFWriter.py

Copyright 2011 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Convert RST to PDF.

"""


import os
import subprocess

from libgutenberg.Logger import debug, info, warning, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat

from ebookmaker import ParserFactory
from ebookmaker import writers
from ebookmaker.CommonCode import Options

options = Options()

class Writer (writers.BaseWriter):
    """ Class to write PDF. """

    def build (self, job):
        """ Build PDF file. """

        inputfilename  = job.url
        outputfilename = os.path.join (os.path.abspath(job.outputdir), job.outputfile)

        debug ("Inputfile: %s" % inputfilename)
        info ("Creating PDF file: %s" % outputfilename)

        parser = ParserFactory.ParserFactory.create (inputfilename)

        if not hasattr (parser, 'rst2xetex'):
            warning ('Skipping PDF Output because input mediatype is %s' % parser.mediatype())
            raise SkipOutputFormat

        # Brain-dead xetex doesn't understand unix pipes
        # so we have to write a temp file

        texfilename = os.path.splitext (outputfilename)[0] + '.tex'
        auxfilename = os.path.splitext (outputfilename)[0] + '.aux'
        logfilename = os.path.splitext (outputfilename)[0] + '.log'

        try:
            os.remove (auxfilename)
        except OSError:
            pass

        tex = parser.rst2xetex (job)
        with open (texfilename, 'wb') as fp:
            fp.write (tex)

        try:
            cwd = os.getcwd ()
            os.chdir (os.path.abspath(job.outputdir))

            _xetex = subprocess.Popen ([options.config.XELATEX,
                                        "-output-directory", job.outputdir,
                                        "-interaction", "nonstopmode",
                                        texfilename],
                                       stdin = subprocess.PIPE,
                                       stdout = subprocess.PIPE,
                                       stderr = subprocess.PIPE)
        except OSError as what:
            os.chdir (cwd)
            error ("PDFWriter: %s %s" % (options.config.XELATEX, what))
            raise SkipOutputFormat

        (dummy_stdout, dummy_stderr) = _xetex.communicate ()

        with open (logfilename, encoding='utf-8') as fp:
            for line in fp:
                line = line.strip ()
                if 'Error:' in line:
                    error ("xetex: %s" % line)
                if options.verbose >= 1:
                    if 'Warning:' in line:
                        warning ("xetex: %s" % line)

        if options.verbose < 2:
            try:
                os.remove (texfilename)
                os.remove (logfilename)
                os.remove (auxfilename)
            except OSError:
                pass

        os.chdir (cwd)

        info ("Done PDF file: %s" % outputfilename)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

Parser Package

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""

from __future__ import unicode_literals

import re
import os
import cchardet
from cherrypy.lib import httputil
import six
from six.moves import urllib

import lxml.html
from lxml import etree
from lxml.builder import ElementMaker

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import NS, xpath
from libgutenberg.MediaTypes import mediatypes as mt
from libgutenberg.Logger import info, debug, warning, error

from ebookmaker.CommonCode import Options

options = Options()

BROKEN = 'resource://ebookmaker.parsers/broken.png'

RE_GUTENBERG = re.compile(r'Project Gutenberg', re.I)
RE_AUTHOR = re.compile(r"^Author:\s+(.+)$", re.I | re.M)
RE_TITLE = re.compile(r"^Title:\s+(.+)$", re.I | re.M)
RE_LANGUAGE = re.compile(r"^Language:\s+(.+)$", re.I | re.M)
# Release Date: September 5, 2009 [EBook #29915]
RE_RELEASEDATE = re.compile(r"^(Release|Posting)\s+Date:\s+(.+)\[", re.I | re.M)
RE_EBOOKNO = re.compile(r'\[E(?:Book|Text) #(\d+)\]', re.I | re.M)

REB_XML_CHARSET = re.compile(br'^<\?xml.*encoding\s*=\s*["\']([^"\'\s]+)', re.I)
RE_HTML_CHARSET = re.compile(r';\s*charset\s*=\s*([^"\'\s]+)', re.I)
REB_HTML_CHARSET = re.compile(br';\s*charset\s*=\s*([^"\'\s]+)', re.I)
REB_PG_CHARSET = re.compile(br"^Character Set Encoding:\s+([-\w\d]+)\s*$", re.I | re.M)


# XML 1.1 RestrictedChars
# [#x1-#x8] | [#xB-#xC] | [#xE-#x1F] | [#x7F-#x84] | [#x86-#x9F]
RE_RESTRICTED = re.compile('[\x01-\x08\x0b\x0c\x0e-\x1f\x7f-\x84\x86-\x9f]')

XML_NAMESTARTCHAR = ':A-Z_a-z\u00c0-\u00d6\u00d8-\u00f6\u00f8-\u02ff' \
                    '\u0370-\u037d\u037f-\u1fff\u200c-\u200d\u2070-\u218f' \
                    '\u2c00-\u2fef\u3001-\ud7ff\uf900-\ufdcf\ufdf0-\ufffd'
                    # u'\U00010000-\U000effff'
XML_NAMECHAR = '-.0-9\u00b7\u0300-\u036f\u203f-\u2040' + XML_NAMESTARTCHAR

RE_XML_NAME = re.compile('^[%s][%s]*$' % (XML_NAMESTARTCHAR, XML_NAMECHAR))

URI_MARK_CHARS = "-_.!~*'()"
URI_RESERVED_CHARS = ';/?:@&=+$,'

RE_URI_FRAGMENT = re.compile('[' + URI_MARK_CHARS + URI_RESERVED_CHARS + '%A-Za-z0-9]+')

# all bogus encoding names used in PG go in here
BOGUS_CHARSET_NAMES = {'iso-latin-1': 'iso-8859-1',
                       'big5': 'big5hkscs',
                       'big-5': 'big5hkscs',

                       # python has bogus codec name
                       'macintosh': 'mac_roman',
                       }

IMAGE_WRAPPER = """<?xml version="1.0"?>{doctype}
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <title>{title}</title>
    <link href="pgepub.css" rel="stylesheet"/>
  </head>
  <body>
    <div style="text-align: center">
      <img src="{src}" alt={title} class="{wrapper_class}" />
      {backlink}
    </div>
  </body>
</html>"""

# exported
em = ElementMaker(makeelement=lxml.html.xhtml_parser.makeelement,
                  namespace=str(NS.xhtml),
                  nsmap={None: str(NS.xhtml)})

def webify_url(url):
    """ make the url for a parser, accounting for platform """
    if re.search(r'^https?://', url):
        return url
    if url.startswith('file:'):
        return url
    if url.startswith('resource:'):
        return url
    if re.search(r'^[a-zA-z]:', url):
        return 'file:///' + url.replace(os.path.sep, '/')
    url = os.path.abspath(url).replace(os.path.sep, '/')
    if url.startswith('/'):
        url = url[1:]
    return 'file:///' + url



class ParserAttributes(object): # pylint: disable=too-few-public-methods
    """ Object to hold attributes for the lifetime of a parser.

    Typical attributes held here would be:
      - url
      - orig_url
      - mediatpye
      - orig_mediatpye
      - referrer
      - id

    mediatype and orig_mediatype are of type HeaderElement

    mediatype is not necessarily the same as orig_mediatype,
    because parsers may convert between input and output, eg.
    reST to xhtml.

    """

    HeaderElement = httputil.HeaderElement

    def __init__(self):
        self.url = None
        self.orig_url = None
        self.mediatype = None
        self.orig_mediatype = None
        self.id = None
        self.rel = set()
        self.referrer = None
        self.title = None


    def update(self, more_attribs):
        for k, v in vars(more_attribs).items():
            if v:
                setattr(self, k, v)


    def get(self, name, default=None):
        return getattr(self, name, default)

    def __str__(self):
        a = []
        for k, v in vars(self).items():
            a.append('%s=%s' % (k, v))
        return '\n'.join(a)


class ParserBase(object):
    """ Base class for more specialized parsers. """

    def __init__(self, attribs=None):
        self.attribs = attribs or ParserAttributes()
        self.attribs.mediatype = self.attribs.orig_mediatype
        self.fp = None
        self.buffer = None
        self.unicode_buffer = None


    def pre_parse(self):
        """ Do a lightweight parse, that allows iterlinks () to succeed.

        Spider.py needs to use iterlinks () to grab dependencies, but
        does not need a full parse. If a lightweight parse doesn't
        make sense, you may also do a full parse here and save the
        result.

        """
        pass


    def parse(self):
        """ Do a full parse.

        When this gets called, a pre_parse has already been done,
        so you might safely reuse any cached results.

        """

        pass


    def get_charset_from_content_type(self):
        """ Get charset from server content-type. """

        charset = self.attribs.orig_mediatype.params.get('charset')
        if charset:
            debug('Got charset %s from server' % charset)
            return charset
        return None


    def get_charset_from_meta(self): # pylint: disable=R0201
        """ Parse header metadata for charset.

        Header metadata can be either of:

          - html metadata
          - rst enconding comment
          - pg header charset line

        Override this as required.

        """

        return None


    def guess_charset_from_body(self):
        """ Guess charset from text. """

        result = cchardet.detect(self.bytes_content())
        charset = result.get('encoding')
        if charset:
            debug('Got charset %s from text sniffing' % charset)
            return charset
        return None


    def bytes_content(self):
        """ Get document content as raw bytes. """

        if self.buffer is None:
            if self.fp is None:
                return b''
            try:
                debug("Fetching %s ..." % self.attribs.url)
                self.buffer = self.fp.read()
                self.fp.close()
            except IOError as what:
                error(what)

        return self.buffer


    def unicode_content(self):
        """ Get document content as unicode string. """

        if self.unicode_buffer is None:
            data = (self.decode(self.get_charset_from_content_type()) or
                    self.decode(self.get_charset_from_meta()) or
                    self.decode(self.guess_charset_from_body()) or
                    self.decode('utf-8') or
                    self.decode('windows-1252'))

            if not data:
                if data == '':
                    info('Continuing parse despite missing file')
                    self.unicode_buffer = ''
                else:
                    raise UnicodeError("Text in Klingon encoding ... giving up.")

            # normalize line-endings
            if '\r' in data or '\u2028' in data:
                data = '\n'.join(data.splitlines())
            self.unicode_buffer = data

        return self.unicode_buffer


    def decode(self, charset):
        """ Try to decode document contents to unicode. """
        if charset is None:
            return None

        charset = charset.lower().strip()

        if charset in BOGUS_CHARSET_NAMES:
            charset = BOGUS_CHARSET_NAMES[charset]

        if charset == 'utf-8':
            charset = 'utf_8_sig'

        try:
            debug("Trying to decode document with charset %s ..." % charset)
            buffer = self.bytes_content()
            buffer = REB_PG_CHARSET.sub(b'', buffer)
            buffer = buffer.decode(charset)
            self.attribs.orig_mediatype.params['charset'] = charset
            return buffer
        except LookupError as what:
            # unknown charset,
            error("Invalid charset name: %s (%s)" % (charset, what))
        except UnicodeError as what:
            # mis-stated charset, did not decode
            error("Text not in charset %s (%s)" % (charset, what))
        return None


    def mediatype(self):
        """ Return parser output mediatype. Helper function. """
        if self.attribs.mediatype:
            return self.attribs.mediatype.value
        return None


    # Links are found in HTMLParserBase and CSSParser. These methods
    # are overwritten there.

    def iterlinks(self): # pylint: disable=R0201
        """ Return all links in document.

        returns a list of url, dict
        dict may contain any of: tag, id, rel, type.

        """

        return []


    def rewrite_links(self, dummy_f): # pylint: disable=R0201
        """ Rewrite all links using the function f. """
        return


    def remap_links(self, dummy_url_map): # pylint: disable=R0201
        """ Rewrite all links using the dictionary url_map. """
        return

class TxtParser(ParserBase):
    """ Base class for text files we don't want to convert.
    """

    def __init__(self, attribs=None):
        ParserBase.__init__(self, attribs)
        self.attribs.mediatype = ParserAttributes.HeaderElement(mt.txt)

    def serialize(self):
        return bytes(self.unicode_content(), 'utf8')

class HTMLParserBase(ParserBase):
    """ Base class for more HTMLish parsers.

    (HTMLParser, GutenbergTextParser)

    """

    def __init__(self, attribs=None):
        ParserBase.__init__(self, attribs)
        self.attribs.mediatype = ParserAttributes.HeaderElement(mt.xhtml)
        self.xhtml = None


    @staticmethod
    def add_class(elem, class_):
        """ Add a class to html element. """
        classes = elem.get('class', '').split()
        classes.append(class_)
        elem.set('class', ' '.join(classes))


    def get_charset_from_meta(self):
        """ Parse text for hints about charset. """

        charset = None
        html = self.bytes_content()

        match = REB_XML_CHARSET.search(html)
        if match:
            charset = match.group(1).decode('ascii')
            debug('Got charset %s from xml declaration' % charset)
        else:
            match = REB_HTML_CHARSET.search(html)
            if match:
                charset = match.group(1).decode('ascii')
                debug('Got charset %s from html meta' % charset)

        return charset


    def iterlinks(self):
        """ Return all links in document. """

        # To keep an image even in non-image build specify
        # class="x-ebookmaker-important"

        keeps = xpath(self.xhtml, "//img[contains (@class, 'x-ebookmaker-important')]")
        for keep in keeps:
            keep.set('rel', 'important')

        # iterate links

        for (elem, dummy_attribute, url, dummy_pos) in self.xhtml.iterlinks():
            yield url, elem


    def rewrite_links(self, f):
        """ Rewrite all links using the function f. """
        self.xhtml.rewrite_links(f)


    def remap_links(self, url_map):
        """ Rewrite all links using the dictionary url_map. """
        def f(url):
            """ Remap function """
            ur, frag = urllib.parse.urldefrag(url)
            if ur in url_map:
                debug("Rewriting redirected url: %s to %s" % (ur, url_map[ur]))
                ur = url_map[ur]
            return "%s#%s" % (ur, frag) if frag else ur

        self.rewrite_links(f)


    @staticmethod
    def strip_links(xhtml, manifest):
        """ Strip all links to urls not in manifest.

        This includes  <link href> and <img src> and, if strip_links is set, <a href>.
        Assume links and urls are already made absolute.

        """

        if options.strip_links:
            for link in xpath(xhtml, '//xhtml:a[@href]'):
                href = urllib.parse.urldefrag(link.get('href'))[0]
                if href not in manifest:
                    debug("strip_links: Deleting <a> to %s not in manifest." % href)
                    del link.attrib['href']

        for link in xpath(xhtml, '//xhtml:link[@href]'):
            href = link.get('href')
            if href not in manifest:
                debug("strip_links: Deleting <link> to %s not in manifest." % href)
                link.drop_tree()

        for image in xpath(xhtml, '//xhtml:img[@src]'):
            src = image.get('src')
            if src not in manifest:
                debug("strip_links: Deleting <img> with src %s not in manifest." % src)
                image.tail = image.get('alt', '') + (image.tail or '')
                image.drop_tree()


    def make_toc(self, xhtml):
        """ Build a TOC from HTML headers.

        Return a list of tuples (url, text, depth).

        Page numbers are also inserted because DTBook NCX needs the
        play_order to be sequential.

        """

        def id_generator(i=0):
            """ Generate an id for the TOC to link to. """
            while True:
                yield 'pgepubid%05d' % i
                i += 1

        def get_header_text(header):
            """ clean header text """
            text = gg.normalize(etree.tostring(header,
                                               method="text",
                                               encoding=six.text_type,
                                               with_tail=False))

            return header.get('title', text).strip()

        idg = id_generator()

        def get_id(elem):
            """ Get the id of the element or generate and set one. """
            if not elem.get('id'):
                elem.set('id', six.next(idg))
            return elem.get('id')

        toc = []

        for header in xpath(
                xhtml,
                '//xhtml:h1|//xhtml:h2|//xhtml:h3|//xhtml:h4|'
                # DP page number
                '//xhtml:*[contains (@class, "pageno")]|'
                # DocUtils contents header
                '//xhtml:p[contains (@class, "topic-title")]'
            ):

            previous = header.getprevious()
            if previous is not None and previous.tag == header.tag:
                # consecutive headers get combined
                continue

            text = get_header_text(header)

            if not text:
                # so <h2 title=""> may be used to suppress TOC entry
                continue

            if header.get('class', '').find('pageno') > -1:
                toc.append(["%s#%s" % (self.attribs.url, get_id(header)), text, -1])
            else:
                # header
                if ((text.lower().startswith('by ')
                     or text.lower() == ('by'))
                        and 'x-ebookmaker-important' not in header.get('class', '')):
                    # common error in PG: <h2>by Lewis Carroll</h2> should
                    # yield no TOC entry
                    warning('dropping by-heading in %s: %s', self.attribs.url, text.lower())
                    continue

                try:
                    depth = int(header.tag[-1:])
                except ValueError:
                    depth = 2 # avoid top level

                #join consecutive headers
                next = header.getnext()
                while next is not None and next.tag == header.tag:
                    text = (text + ' ' + get_header_text(next)).strip()
                    next = next.getnext()

                # if <h*> is first element of a <div> use <div> instead
                parent = header.getparent()
                if (parent.tag == NS.xhtml.div and
                        parent[0] == header and
                        parent.text and
                        parent.text.strip() == ''):
                    header = parent

                toc.append(["%s#%s" % (self.attribs.url, get_id(header)), text, depth])

        return toc


    def serialize(self):
        """ Serialize to string. """
        if self.attribs.url.endswith('.xhtml'):
            # epub3
            doctype = ""
        else:
            # epub2
            doctype=gg.XHTML_DOCTYPE
        return etree.tostring(self.xhtml,
                              doctype=doctype,
                              xml_declaration=True,
                              encoding='utf-8',
                              pretty_print=True)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""

Module parsers

Copyright 2010-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Customized Project Gutenberg directives for RST parser.

"""

from docutils import statemachine
from docutils.parsers.rst import Directive, directives

from ebookmaker.mydocutils import parsers

from ebookmaker.mydocutils.gutenberg import transforms as gutenberg_transforms

from libgutenberg.Logger import error, warning, info, debug

# pylint: disable=W0142, W0102


class PGHeaderFooter (Directive):
    """ Inserts PG header or footer. """

    required_arguments = 0
    optional_arguments = 0

    def run (self):
        settings = self.state.document.settings
        include_lines = statemachine.string2lines (
            settings.get_resource ('mydocutils.gutenberg.parsers', self.resource),
            settings.tab_width,
            convert_whitespace = 1)
        self.state_machine.insert_input (include_lines, '')
        return []


class PGHeader (PGHeaderFooter):
    """ Inserts PG header. """
    resource = 'pg-header.rst'


class PGFooter (PGHeaderFooter):
    """ Inserts PG footer. """
    resource = 'pg-footer.rst'


class Parser (parsers.Parser):
    """ Parser with PG custom directives. """

    def __init__ (self):
        parsers.Parser.__init__ (self)

        directives.register_directive ('pgheader',        PGHeader)
        directives.register_directive ('pgfooter',        PGFooter)


    def get_transforms (self):
        return parsers.Parser.get_transforms (self) + [
            gutenberg_transforms.VariablesTransform,
            gutenberg_transforms.SubRefToVarTransform]



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

ImageParser.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Parse an url of type image/*.

"""

import copy

import six
from PIL import Image

from pkg_resources import resource_stream # pylint: disable=E0611

from libgutenberg.Logger import debug, error
from libgutenberg.MediaTypes import mediatypes as mt
from ebookmaker.parsers import ParserBase
from ebookmaker.ParserFactory import ParserFactory

mediatypes = (mt.jpeg, mt.png, mt.gif, mt.svg)

class Parser(ParserBase):
    """Parse an image.

    And maybe resize it for ePub packaging.

    """

    def __init__(self, attribs=None):
        ParserBase.__init__(self, attribs)
        self.image_data = None
        self.dimen = None


    def resize_image(self, max_size, max_dimen, output_format=None):
        """ Create a new parser with a resized image. """

        def scale_image(image, scale):
            was = ''
            if scale < 1.0:
                dimen = (int(image.size[0] * scale), int(image.size[1] * scale))
                was = "(was %d x %d scale=%.2f) " % (image.size[0], image.size[1], scale)
                image = image.resize(dimen, Image.ANTIALIAS)
            return was, image

        def get_image_data(image, format_, quality='keep'):
            """ Format is the output format, not necessarily the input format """
            buf = six.BytesIO()
            if image.format != 'JPEG' and quality == 'keep':
                quality = 90
            if format_ == 'png':
                image.save(buf, 'png', optimize=True)
            else:
                try:
                    image.save(buf, 'jpeg', quality=quality)
                except ValueError as e:
                    if quality == 'keep' and 'quantization' in str(e):
                        image.save(buf, 'jpeg', quality=90)
                    else:
                        raise e
            return buf.getvalue()
        
        # can't do anything with SVG files
        if self.attribs.url.endswith('.svg'):
            return self

        new_parser = Parser()

        try:
            unsized_image = Image.open(six.BytesIO(self.image_data))

            format_ = unsized_image.format.lower()
            if output_format:
                format_ = output_format
            if format_ == 'gif':
                format_ = 'png'
            if format_ == 'jpeg' and unsized_image.mode.lower() not in ('rgb', 'l'):
                unsized_image = unsized_image.convert('RGB')

            if 'dpi' in unsized_image.info:
                del unsized_image.info['dpi']

            # maybe resize image

            # find scaling factor
            scale = 1.0
            scale = min(scale, max_dimen[0] / float(unsized_image.size[0]))
            scale = min(scale, max_dimen[1] / float(unsized_image.size[1]))

            was, image = scale_image(unsized_image, scale)
            data = get_image_data(image, format_)

            if format_ == 'png':
                # scale it till it fits into max_size
                while len(data) > max_size and scale > 0.01:
                    scale = scale * 0.8
                    was, image = scale_image(unsized_image, scale)
                    data = get_image_data(image, format_)
            else:
                # find best quality that fits into max_size
                if len(data) > max_size:
                    for quality in (90, 85, 80, 70, 60, 50, 40, 30, 20, 10):
                        data = get_image_data(image, format_, quality=quality)
                        if len(data) <= max_size:
                            break

                    was += 'q=%d' % quality
            comment = "Image: %d x %d size=%d %s" % (
                image.size[0], image.size[1], len(data), was
            )
            debug(comment)

            new_parser.image_data = data
            new_parser.dimen = tuple(image.size)

            new_parser.attribs = copy.copy(self.attribs)
            new_parser.attribs.comment = comment
            new_parser.fp = self.fp

        except IOError as what:
            error("Could not resize image: %s", self.attribs.url)
            new_parser.attribs = copy.copy(self.attribs)
            fp = resource_stream('ebookmaker.parsers', 'broken.png')
            new_parser.image_data = fp.read()
            fp.close()

        return new_parser


    def get_image_dimen(self):
        if self.dimen is None:
            if self.image_data:
                try:
                    image = Image.open(six.BytesIO(self.image_data))
                    self.dimen = image.size
                except IOError as what:
                    error("Could not resize image (probably broken): %s", self.attribs.url)
                    self.dimen = (0, 0)  # broken image
            else:
                self.dimen = (0, 0)  # broken image
        return self.dimen


    def pre_parse(self):
        if self.image_data is None:
            self.image_data = self.bytes_content()

    def parse(self):
        pass

    def serialize(self):
        """ Serialize the image. """
        return self.image_data



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

KindleWriter.py

Copyright 2009-2012 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

"""

import re
import os
import subprocess
import sys

from libgutenberg.Logger import info, debug, warning, error
from libgutenberg.GutenbergGlobals import SkipOutputFormat
from ebookmaker.writers import BaseWriter
from ebookmaker.CommonCode import Options

options = Options()
no_kindlegen_langs = ['ceb', 'eo', 'fur', 'ia', 'ilo', 'iu', 'mi',
                      'myn', 'nah', 'nap', 'oc', 'oji', 'tl', 'bo']

class Writer(BaseWriter):
    """ Class for writing kindle files. """


    def build(self, job):
        """ Build kindle file from epub using amazon kindlegen or calibre. """

        if job.maintype == 'kindle':
            if job.dc.languages:
                if job.dc.languages[0].id in no_kindlegen_langs:
                    mobimaker = options.config.MOBILANG
                else:
                    mobimaker = options.config.MOBIGEN
        else:
            mobimaker = options.config.MOBIKF8
        if not mobimaker:
            info('no mobimaker available')
            return

        # kindlegen needs localized paths
        outputdir = os.path.abspath(job.outputdir)

        info("Creating Kindle file: %s" % os.path.join(outputdir, job.outputfile))
        info("            ... from: %s" % job.url)

        try:
            cwd = os.getcwd()
            os.chdir(outputdir)
            if 'ebook-convert' in mobimaker:
                kindlegen = subprocess.Popen(
                    [
                        mobimaker,
                        job.url,
                        os.path.basename(job.outputfile),
                        '--personal-doc="[EBOK]"',
                        '--mobi-file-type=' + ('new' if job.maintype == 'kf8' else 'old')
                    ],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )
            else:
                kindlegen = subprocess.Popen(
                    [
                        mobimaker,
                        '-o', os.path.basename(job.outputfile),
                        job.url
                    ],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE
                )

        except OSError as what:
            os.chdir(cwd)
            error("KindleWriter: %s %s" % (mobimaker, what))
            raise SkipOutputFormat

        (stdout, stderr) = kindlegen.communicate()

        os.chdir(cwd)

        if kindlegen.returncode > 0:
            regex = re.compile(r'^(\w+)\(prcgen\):')

            # pylint: disable=E1103
            msg = stderr.rstrip()
            if msg:
                msg = msg.decode(sys.stderr.encoding)
                error(msg)
            msg = stdout.rstrip()
            msg = msg.decode(sys.stdout.encoding)
            for line in msg.splitlines():
                match = regex.match(line)
                if match:
                    sline = regex.sub("", line)
                    g = match.group(1).lower()
                    if g == 'info':
                        if sline == 'MOBI File generated with WARNINGS!':
                            # we knew that already
                            continue
                        # info("kindlegen: %s" % sline)
                    elif g == 'warning':
                        if sline.startswith('Cover is too small'):
                            continue
                        if sline == 'Cover not specified':
                            continue
                        warning("kindlegen: %s" % sline)
                    elif g == 'error':
                        error("kindlegen: %s" % sline)
                    else:
                        error(line)

        info("Done Kindle file: %s" % os.path.join(outputdir, job.outputfile))
